import os
import re
import sys
import json
from datetime import datetime as dt
import zipfile
import string
import argparse
import traceback
import functools
import itertools
import subprocess as sb
from pathlib import Path

MAX_FILE_SIZE = 1024 * 1024  # 1MB
MANIFEST_OLD_PATH = "META-INF/MANIFEST.MF"

def take(n, l):
    for i, item in enumerate(l):
        if i > n:
            break
        yield item


def wrap_in(exctype, msg):
    def inner(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except BaseException as e:
                raise exctype(msg) from e

        return wrapper

    return inner


class Jar:
    def __init__(self, path):
        self.path = path
        self._manifest = {}
        self._dirlist = []

    def _parse_manifest(self, lines):
        version_indication = "version="
        old_version_indication = "-Version"
        version_lines = [line for line in lines if line.startswith(version_indication)]

        if len(version_lines) > 0:
            version = version_lines[0][len(version_indication):]
            yield 'Version', version.strip()

        version_lines = [line for line in lines if old_version_indication in line]
        for ver_line in version_lines:
            key, value = ver_line.split(':')
            yield key.strip(), value.strip()

    def _open(self):
        if not zipfile.is_zipfile(self.path):
            raise ValueError("path is not a zip file: {}".format(self.path))
        return zipfile.ZipFile(self.path)

    def _read_dirlist(self):
        with self._open() as zf:
            filenames = {p for p in zf.namelist()}
            return [f for f in filenames if any(r.search(f.lower()) for r in args.dirlist)]



    def _get_manifest_path(self, zf):
        for path in [args.manifest_path, MANIFEST_OLD_PATH]:
            if path in zf.namelist():
                return path

    def _read_manifest(self, throw_on_error=False):
        try:
            with self._open() as zf:
                manifest_path = self._get_manifest_path(zf)
                if not manifest_path:
                    # Not found manifest file
                    return {}

                manifest_info = zf.getinfo(manifest_path)
                if manifest_info.file_size > MAX_FILE_SIZE:
                    raise IOError("manifest file is too big")

                with zf.open(manifest_path) as f:
                    readline_f = functools.partial(f.readline, MAX_FILE_SIZE)
                    manifest_lines = list(x.decode().strip() for x in iter(readline_f, b''))
                    manifest = self._parse_manifest(manifest_lines)
                    return {k: v for k, v in manifest
                            if not args.manifest_keys or any(m.search(k.lower()) for m in args.manifest_keys)}
        except:
            print("error while reading manifest of '{}': {}".format(
                self.path, traceback.format_exc()), file=sys.stderr)

            if throw_on_error:
                raise

            return {}

    def manifest(self, throw_on_error=False):
        if not self._manifest:
            self._manifest = self._read_manifest(throw_on_error)
        return self._manifest

    def dirlist(self):
        if not self._dirlist:
            self._dirlist = self._read_dirlist()
        return self._dirlist


class FatalArgumentError(Exception):
    pass


class State:
    def __init__(self, name, payload):
        self.payload = payload
        self.merge_cached_content = args.merge_cached_content
        self.path = self._get_location(name)

    @staticmethod
    def remove(name):
        path = State._get_location(name)
        if path.exists():
            path.unlink()
            return path

    @staticmethod
    def _resources_dir():
        for path in map(Path,
                    ["/opt/microsoft/mdatp/resources",
                     "/Applications/Microsoft Defender.app/Contents/MacOS/wdavdaemon_enterprise.app/Contents/Resources",
                     "/Applications/Microsoft Defender ATP.app/Contents/MacOS/wdavdaemon_enterprise.app/Contents/Resources"]):
            if path.exists():
                return path

        raise ValueError("platform not supported: {}".format(sys.platform))

    @staticmethod
    def _get_location(name):
        basedir = State._resources_dir() / "cache"
        if not basedir.exists():
            basedir.mkdir()
        return (basedir / name).with_suffix(".json")

    def __enter__(self):
        if not self.path.exists():
            return

        if not self.merge_cached_content:
            return

        with open(self.path, 'r') as f:
            old_payload = json.load(f)

        current_handles = {h['name'] for h in self.payload['handlers']}

        for handler in old_payload['handlers']:
            name, pid = handler['name'], handler['pid']

            if name in current_handles or not Path(name).exists():
                continue

            handler['active'] = False
            self.payload['handlers'].append(handler)
            self.payload['environ'][pid] = old_payload['environ'][pid]
            self.payload['manifest'][name] = old_payload['manifest'][name]
            self.payload['dirlist'][name] = old_payload['dirlist'][name]
            self.payload['ports'][name] = old_payload['ports'][name]

    def __exit__(self, *args, **kwargs):
        with open(self.path, 'w') as f:
            json.dump(self.payload, f, separators=(',', ':'))


def validate_state_id(dest):
    if not re.match('^[A-Za-z0-9._]+$', dest):
        raise ValueError("state id invalid: {}".format(dest))
    return dest


def validate_command_name(dest):
    commands = [x.strip() for x in dest.split(",")]
    for command in commands:
        if not re.match('^[A-Za-z0-9]+$', command):
            raise ValueError("command name invalid: {}".format(command))
    return commands


def split_by_comma(dest):
    return [x.strip() for x in dest.split(",")]


def split_and_compile(dest):
    return [re.compile(r) for r in split_by_comma(dest)]


def validate_limit(dest):
    limit = int(dest)
    if limit < 1:
        raise ValueError("limit is too small")
    return limit


@wrap_in(FatalArgumentError, "unable to parse arguments")
def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--ScriptName", dest='script_name')
    parser.add_argument("--id", type=validate_state_id)
    parser.add_argument("--limit", type=validate_limit,
                        default=100, help='limit results. defaults to 100.')
    parser.add_argument("--limit-files", type=validate_limit,
                        default=2000, help='limit results. defaults to 2000.')
    parser.add_argument("--filter-command", type=validate_command_name,
                        dest='lsof_commands', help='commands to filter')
    parser.add_argument("--filter-name", type=split_and_compile,
                        default=[], help='filter handler name')
    parser.add_argument("--filter-env", type=split_and_compile,
                        default=[], help='filter env vars')
    parser.add_argument("--manifest-path", help='manifest path to read')
    parser.add_argument("--marker-path", dest='mitigation_marked', type=lambda path: Path(path).exists(),
                        default=False,  help='check existence of marker file path')
    parser.add_argument("--collect-manifest-keys", dest='manifest_keys',
                        type=split_and_compile, default=[], help='filter manifest keys')
    parser.add_argument("--collect-dirlist", dest='dirlist',
                        type=split_and_compile, default=[], help='filter filenames')
    parser.add_argument("--merge-cached-content", type=bool, default=True)
    parser.add_argument("--remove", type=validate_state_id,
                        help="remove a specific state, by id")

    args = parser.parse_args()
    if not args.remove and any([not args.id, not args.lsof_commands]):
        parser.error("you must pass --id and --filter-command")
    return args


def _run_cmd(cmd):
    with sb.Popen(cmd, stdout=sb.PIPE, bufsize=1, universal_newlines=True, shell=True) as p:
        yield from map(str.strip, p.stdout)


def grab_environ(pid):
    # use 'ps' to grab the env variables of given pid
    cmd = "ps eww -o command {} | tr ' ' '\n' | tail -n+3".format(int(pid))
    for line in _run_cmd(cmd):
        if not args.filter_env or any(x.search(line) for x in args.filter_env):
            yield line


def grab_ports(pid):
    for line in list_open_file_handles('-nPi', '-p {}'.format(pid)):
        yield dict(type=line['type'], proto=line['node'], addr=line['name'])


def is_zip_supported_format(name):
    return name.endswith('.jar') or name.endswith('.war') or name.endswith('.ear')


def list_open_file_handles(*flags):
    # run lsof with given parameters, take only first N handles (ignore the rest)
    cmd = 'lsof {} -a -c {} 2>/dev/null'.format(' '.join(flags), ' -c '.join(args.lsof_commands))
    handles_gen = _run_cmd(cmd)

    # if lsof returned output, the first line would be the column names
    column_names = next(handles_gen, "").strip().lower().split()
    if not column_names:
        return

    # iterate each handler & parse it
    for handler_line in handles_gen:
        ts = dt.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')
        handler = dict(zip(column_names, handler_line.split()),
                       timestamp=ts, active=True)
        yield handler


def lsof_entries():
    # iterate each handler & parse it
    for handler in list_open_file_handles():
        handler_name = handler.get('name')

        # handler name must exist
        if not handler_name:
            continue

        yield handler


def capping_lsof_entries(handles_list):
    if len(handles_list) > args.limit_files:
        required_handles = []
        handles_to_check_dirlist = []
        for h in handles_list:
            if not args.filter_name or any(x.search(h['name'].lower()) for x in args.filter_name):
                required_handles.append(h)
            else:
                handles_to_check_dirlist.append(h)

        # we prefer to give priority to files we need according to filter-name arg and then we add the files to check for nested jars
        if len(required_handles) > args.limit_files:
            return take(args.limit_files, required_handles)
        else:
            free_space = args.limit_files - len(required_handles)
            required_handles.append(list(take(free_space, handles_to_check_dirlist)))
            return required_handles
    else:
        return handles_list


@wrap_in(ValueError, "runtime error")
def main():
    if os.geteuid() != 0:
        raise ValueError("script needs to be run as root")

    if args.remove:
        path = State.remove(args.remove)
        if not path:
            raise OSError("state file doesn't exist: {}".format(args.remove))

    # keep all file handles we need and files that contains jars we need (ubar-jars)
    handles = capping_lsof_entries(list(lsof_entries()))
    jars = {h['name']: Jar(h['name']) for h in handles if is_zip_supported_format(h['name'])}
    dirlist = {name: jar.dirlist() for name, jar in jars.items() if jar.dirlist()}

    # keep only relevant handles (the required file or contains a required files)
    filtered_handles = []
    for h in handles:
        # required file handle name
        if not args.filter_name or any(x.search(h['name'].lower()) for x in args.filter_name):
            filtered_handles.append(h)
            continue

        # required nested jars (we filter specific keys according to collect-dirlist arg)
        if h['name'] in dirlist:
            filtered_handles.append(h)

    filtered_jars = {h['name']: Jar(h['name']) for h in filtered_handles}
    pids = set(h['pid'] for h in filtered_handles)
    environ = {pid: list(grab_environ(pid)) for pid in pids}
    ports = {pid: list(take(args.limit, grab_ports(pid))) for pid in pids}

    # read only manifest of file names according to args.filter_name
    manifest = {name: jar.manifest() for name, jar in filtered_jars.items() if
                any(x.search(name.lower()) for x in args.filter_name) and jar.manifest()}

    mitigation_marked = args.mitigation_marked

    payload = dict(scriptVersion=2,
                   cmdline=" ".join(sys.argv[1:]),
                   handlers=filtered_handles,
                   ports=ports,
                   environ=environ,
                   dirlist=dirlist,
                   manifest=manifest,
                   mitigation_marked=mitigation_marked)

    with State(name=args.id, payload=payload):
        print(json.dumps(payload), file=sys.stdout)


if __name__ == "__main__":
    try:
        args = parse_args()
        sys.exit(main())
    except FatalArgumentError as e:
        sys.exit(2)
    except BaseException as e:
        if isinstance(e, SystemExit) and e.code in (0, None):
            sys.exit(0)
        print(traceback.format_exc(), file=sys.stderr)
        sys.exit(1)
